/* Copyright (C) 2004 Bart
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include "../include/Device.h"

/*
 *	This is ridiculous!, if i use this one as a member function realpath starts complaining and won't compile :-S
 */
Glib::ustring get_sym_path( const Glib::ustring & real_path ) 
{ 
	int major,minor,size;
	char temp[4096], device_name[4096],short_path[4096] ;
	
	FILE*	proc_part_file = fopen ("/proc/partitions", "r");
	if (!proc_part_file)
		return real_path;
		
	//skip first 2 useless rules of /proc/partitions
	fgets (temp, 256, proc_part_file);fgets (temp, 256, proc_part_file);

	while (fgets (temp, 4096, proc_part_file) && sscanf (temp, "%d %d %d %255s", &major, &minor, &size, device_name) == 4)
	{
		strcpy(short_path, "/dev/"); strcat( short_path, device_name );
		realpath( short_path, device_name );
	
		if ( real_path == device_name ) { 
			fclose (proc_part_file);
		  	return (Glib::ustring(short_path));
		}
		
	}
	
	//it's very unlikely, if not impossible, the program will ever reach this, but just in case.... ;-)
	fclose (proc_part_file);	
	return real_path;
}

Glib::ustring error_message;
bool show_libparted_message = true ;
PedExceptionOption PedException_Handler (PedException* ex)
{
	error_message = ex -> message ;
	
	if ( show_libparted_message )
	{
		Gtk::Label *label_temp;
		Gtk::HBox *hbox_filler;
		
		Glib::ustring message = "<span weight=\"bold\" size=\"larger\">LibParted Message</span>\n\n" ;
		message += "This message comes directly from libparted and has little to do with GParted. ";
		message += "It is possible this message contains usefull information, so you might decide to read it.\n";
		message += "If you don't want these messages anymore, it's safe to disable them.";
				
		Gtk::MessageDialog dialog(  message,true,Gtk::MESSAGE_WARNING , Gtk::BUTTONS_OK, true); 
		Gtk::Frame frame_libparted_message ;
		frame_libparted_message .set_size_request( 400, -1 );
		label_temp = manage( new Gtk::Label("<b>Libparted message: </b>" ) ) ;
		label_temp ->set_use_markup( true ) ;
		frame_libparted_message .set_label_widget( *label_temp ) ;
		
		frame_libparted_message .set_border_width( 5 ) ;
		label_temp = manage( new Gtk::Label( ) ) ;
		label_temp ->set_size_request( 55, -1 ) ; //used like a simple filler ( remember this html cursus :-P )
		hbox_filler = manage( new Gtk::HBox() ) ;
		hbox_filler ->pack_start ( *label_temp, Gtk::PACK_SHRINK ) ;
		hbox_filler ->pack_start ( frame_libparted_message, Gtk::PACK_SHRINK ) ;
		
		dialog .get_vbox() ->pack_start( *hbox_filler, Gtk::PACK_SHRINK ) ;
		
		Gtk::Label label_libparted_message( "<i>" + error_message + "</i>" ) ;
		label_libparted_message .set_selectable( true ) ;
		label_libparted_message .set_use_markup( true ) ;
		label_libparted_message .set_line_wrap( true ) ;
		frame_libparted_message.add ( label_libparted_message ) ;
			
		Gtk::CheckButton checkbutton_message( "Hide libparted messages for this session" ) ;
		checkbutton_message .set_border_width( 5 ) ;
		label_temp = manage( new Gtk::Label(  ) ) ;
		label_temp ->set_size_request( 55, -1 ) ;
		hbox_filler = manage( new Gtk::HBox() ) ;
		hbox_filler ->pack_start ( *label_temp, Gtk::PACK_SHRINK ) ;
		hbox_filler ->pack_start ( checkbutton_message, Gtk::PACK_SHRINK ) ;
		
		dialog .get_vbox() ->pack_start( *hbox_filler, Gtk::PACK_SHRINK ) ;
		
		dialog .show_all_children() ;
		dialog.run(); 
		
		show_libparted_message = ! checkbutton_message .get_active() ;
	}
		
	return PED_EXCEPTION_UNHANDLED ;
}


namespace GParted
{
	
Device::Device()
{
}
	
Device::Device( const Glib::ustring & device_path )
{
	ped_exception_set_handler( PedException_Handler ) ; //didn't like it ;-)
	
	this ->realpath 	=	device_path ; //this one is used by open_device_and_disk
	
	this ->length = 0;//lazy check.. if something goes wrong while reading the device, length will stay zero and i will know it ( see Win_GParted::Find_Devices )
	
	if ( ! open_device_and_disk() )
		return ;
		
	this ->model 		=	device ->model ;
	this ->path 		=	get_sym_path( device ->path ) ;
	this ->disktype 	= 	disk ->type ->name ;
	this ->heads 		=	device ->bios_geom.heads ;
	this ->sectors 	=	device ->bios_geom.sectors ;
	this ->cylinders	=	device ->bios_geom.cylinders ;
	this ->length 		=	heads * sectors * cylinders ;
	
	//get valid flags for this device
	for ( PedPartitionFlag flag = ped_partition_flag_next ( (PedPartitionFlag) 0 ) ; flag ; flag = ped_partition_flag_next ( flag ) )
		flags .push_back( flag ) ;
	
	Read_Disk_Layout() ; 
	
}

void Device::Read_Disk_Layout()
{
	/*
	Reading all partitions and unallocated space correctly from a device proved quite tricky.
	Atm i use the following algoritm:
	
	-	No partitions? the whole device is unallocated space
	-	loop through all partitionNUMBERS and place them in the order in which they exist on the disk.
	-	loop through all partitions and place PRIMARY and EXTENDED in prim_temp and LOGICAL in ext_temp
	-	loop through both vectors and scan for unallocated space. Add findings to prim_final and ext_final ( thus these two contain both partitions and unallocated space )
	-	insert ext_final *after* the extended partition in prim_final
	- 	return prim_final
	
	Also note i'm only interested in unallocated space with size of at least 1 MB. let me know if you don;t like this!
	
	Of course i do some more checking along the way, but this is the big picture. 
	For who dares troll in here, have fun ;-)
	*/
	
	std::vector<Partition> prim_temp,ext_temp,prim_final,ext_final;
	Sector ext_start=0,ext_end=0;
	Partition partition_temp ;

	//clear partitions
	this->device_partitions.clear () ;
	
	//check if there are any partitions at all...
	if ( ped_disk_get_last_partition_num( disk ) == -1 )
	{
		partition_temp.Set_Unallocated( 0, this ->length -1 , false );
		prim_final.push_back( partition_temp );
		this->device_partitions = prim_final;
		return ;
	}
	
	//place partition numbers in the order in which they exist on the disk (generally they should be just ascending from 1 to <# of partitions>
	//but some layouts are different)
	//on the fly removal of numbers which do not contain a partition
	std::vector<int> partitions;
	
	unsigned int t=0;
	for ( int i= ped_disk_get_last_partition_num( disk ) ;i > 0 ; i-- )
	{

		if (ped_disk_get_partition( disk, i )  )
		{
			for ( t=0; t<partitions.size() ; t++ )
				if  (  ped_disk_get_partition( disk, partitions[t] ) ->geom .start   >  ped_disk_get_partition( disk, i )  ->geom .start ) 
					break;//insert partitionnumber at this t
			
			partitions.insert( partitions.begin() + t, i );
		}
	
	}
	
	Glib::ustring filesystem ;
	//place real partitions in their vectors
	for ( unsigned int t=0;t<partitions.size();t++)
	{
		os << this ->path << partitions[t]  ;
		c_partition = ped_disk_get_partition( disk, partitions[t] ) ;
			
		switch( c_partition ->type )
		{ 
			//NORMAL (PRIMARY)
			case 0 :	if ( c_partition ->fs_type )
								filesystem = c_partition ->fs_type ->name ;
							else
								{filesystem = "unknown" ; this ->error = "Unable to detect filesystem! Possible reasons are:\n-The filesystem is somehow damaged\n-The filesystem is unknown to libparted\n-There is simply no filesystem available (unformatted)"; }
							
							partition_temp.Set( os.str(), partitions[t] , GParted::PRIMARY, filesystem, c_partition ->geom .start, c_partition ->geom .end, Get_Used_Sectors( c_partition , os.str() ) , false, ped_partition_is_busy(  c_partition ) );
																	
							prim_temp.push_back( partition_temp );
							prim_temp.back() .flags = Get_Flags( c_partition ) ;			
							prim_temp.back() .error = this ->error ;
					
							break;	
			//LOGICAL
			case 1:	if ( c_partition ->fs_type )
								filesystem = c_partition ->fs_type ->name ;
							else
								{filesystem = "unknown" ; this ->error = "Unable to detect filesystem! Possible reasons are:\n-The filesystem is somehow damaged\n-The filesystem is unknown to libparted\n-There is simply no filesystem available (unformatted)"; }
								
							partition_temp.Set( os.str(), partitions[t] , GParted::LOGICAL, filesystem, c_partition ->geom .start, c_partition ->geom .end, Get_Used_Sectors( c_partition , os.str() ) , true, ped_partition_is_busy( c_partition ) );
																
							ext_temp.push_back( partition_temp );
							ext_temp.back() .flags = Get_Flags( c_partition ) ;	
							ext_temp.back() .error = this ->error ;
												
							break;
			//EXTENDED
			case 2:	ext_start	=	c_partition ->geom .start ;
							ext_end	=	c_partition ->geom .end  ;
							
							partition_temp.Set( os.str(), partitions[t] , GParted::EXTENDED, "extended", ext_start, ext_end , -1, false, ped_partition_is_busy(  c_partition ) );
											
							prim_temp.push_back( partition_temp );
							prim_temp.back() .flags = Get_Flags( c_partition ) ;			
							prim_temp.back() .error = this ->error ;
				
							break;
			//FREESPACE and METADATA 
			default:	Gtk::MessageDialog dialog( "Please mail me to tell what you've done :^)",false, Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true); dialog.run();	
							break;
		}
			
		//wipe both error messages	
		this ->error = "";
		error_message = "";
		
		os.str("");
	}
	
	//now we detect and insert the unallocated space in each of the two vectors
		
	//prim_temp ( begin device <----> end device )
	//unallocated on begin of prim_temp
	if ( (prim_temp.front().sector_start *512/1024/1024) > 0 )
	{
		partition_temp.Set_Unallocated(  0, prim_temp.front().sector_start -1, false );
		prim_final.push_back( partition_temp );
	}
	//now loop through the other primary partitions
	for ( unsigned int t=0;t<prim_temp.size() -1;t++ )
	{
		prim_final.push_back( prim_temp[t] );//first add partition and then check for unallocated
		if ( (  ( prim_temp[t+1].sector_start - prim_temp[t].sector_end ) *512/1024/1024) > 0 ) 
		{
			partition_temp.Set_Unallocated(  prim_temp[t].sector_end +1,  prim_temp[t+1].sector_start -1, false );
			prim_final.push_back( partition_temp );
		}
	}
	//because looping only goes to size -1 we have to add the last element of prim_temp to prim_final
	prim_final.push_back( prim_temp.back() );
	
	//check  for unallocated at end of prim_temp
	if ( ( (this ->length - prim_final.back().sector_end ) * 512/1024/1024) > 0 )
	{
		partition_temp.Set_Unallocated( prim_final.back().sector_end +1 , this ->length -1 , false );
		prim_final.push_back( partition_temp );
	}
	
	prim_temp.clear();//is no longer needed
	
	
	//ext_temp ( begin extended <----> end extended )
	//first check ( extended space is always a little tricky :^) )
	if ( ext_end - ext_start == 0 ) {
		this->device_partitions = prim_final; return ; } //there is no extended partition, prim_final is diskrepresentation
	
	//second check on the vector ;)
	if ( ext_temp.size() == 0 ) //whole extended partition consists of unallocated space
	{
		partition_temp.Set_Unallocated(  ext_start, ext_end , true );
		ext_final.push_back( partition_temp );
	}
	
	else // at least one logical partition in ext_temp
	{
		
		//unallocated on begin of ext_temp
		if ( ( (ext_temp.front().sector_start - ext_start) *512/1024/1024) > 0 )
		{
			partition_temp.Set_Unallocated(  ext_start, ext_temp.front().sector_start -1 , true );
			ext_final.push_back( partition_temp );
		}
		//now loop through the other logical partitions
		for ( unsigned int t=0;t<ext_temp.size() -1;t++ )
		{
			ext_final.push_back( ext_temp[t] );//first add partition and then check for unallocated
			if ( (  ( ext_temp[t+1].sector_start - ext_temp[t].sector_end ) *512/1024/1024) > 0 ) 
			{
				partition_temp.Set_Unallocated(  ext_temp[t].sector_end +1, ext_temp[t+1].sector_start -1 , true );
				ext_final.push_back( partition_temp );
			}
		}
		//because looping only goes to size -1 we have to add the last element of ext_temp to ext_final
		ext_final.push_back( ext_temp.back() );
		//check  for unallocated at end of ext_temp
		if ( ( (ext_end - ext_final.back().sector_end ) * 512/1024/1024) > 0 )
		{
			partition_temp.Set_Unallocated(  ext_final.back().sector_end +1, ext_end , true );
			ext_final.push_back( partition_temp );
		}
		
	}
	
	//Now we have two vectors: prim_final and ext_final. Inserting ext_final *after* the extended partition in prim_final makes prim_final the
	//perfect visual represenation of the disk
	for ( unsigned int t=0;t<prim_final.size();t++ )
	{
		if ( prim_final[t].type == GParted::EXTENDED )
		{			
			prim_final.insert( prim_final.begin() +t +1, ext_final.begin(),ext_final.end() );
			break;
		}
	}

	//return the final vector
	this->device_partitions = prim_final;

}

bool Device::Delete_Partition( const int part_num ) 
{
	if ( ! ped_disk_delete_partition( disk, ped_disk_get_partition( disk, part_num ) )  )
		return false;
		
	return ped_disk_commit( disk ) ;
}

bool Device::Create_Partition( const Partition & new_part, PedTimer * timer) 
{
	//create constraint
	PedConstraint *constraint = NULL;
	PedFileSystemType *fs_type = NULL;
	
	if ( new_part. type ==GParted::EXTENDED )
		constraint = ped_constraint_any ( device);
	else	 
	{
		fs_type = ped_file_system_type_get ( new_part.filesystem.c_str() ) ;
		
		if ( fs_type )
		   constraint = ped_file_system_get_create_constraint ( fs_type, device );
	}
	
	if ( ! constraint )
		return false;
	
	//add new partition
	PedPartitionType type;
	switch ( new_part .type )
	{
		case 0	:	type = PED_PARTITION_NORMAL; 		break;
		case 1	:	type = PED_PARTITION_LOGICAL; 		break;
		case 2	:	type = PED_PARTITION_EXTENDED; 	break;
		default	:	break;
	}
	
	PedPartition *c_part = ped_partition_new ( disk, type,	fs_type, new_part .sector_start, new_part .sector_end ) ;
	
	if ( ! ped_disk_add_partition (disk, c_part, constraint) )
		return false ;
	
	//add filesystem to new partition
	if ( new_part .type == GParted::EXTENDED )    //no need to create a filesystem on an extended partition
		return ped_disk_commit( disk ) ;
		
	//create a filesystem
	PedFileSystem *fs = NULL;
	fs = ped_file_system_create ( & c_part->geom, c_part ->fs_type, timer );
	if ( ! fs )
		return false; 
	else	
		ped_file_system_close (fs);
	
	return ped_disk_commit( disk ) ;
}

bool Device::Move_Resize_Partition( const Partition & partition , PedTimer *timer)
{
	//since resizing extended is a bit different i handle it elsewhere..
	if ( partition.type == GParted::EXTENDED )
		return Resize_Extended( partition, timer ) ;
	
	//normal partitions....
	PedPartition *c_part =  NULL;
	
	c_part = ped_disk_get_partition( disk, partition .partition_number ) ;
	if ( ! c_part )
		return false;
		
	PedFileSystem *fs = NULL;
	
	fs = ped_file_system_open (& c_part->geom );
	if ( ! fs )
		return false ;
	
	PedConstraint *constraint = NULL;
	
	constraint = ped_file_system_get_resize_constraint (fs);
	if ( ! constraint )
	{
		ped_file_system_close (fs);
		return false;
	}
	
	if ( ! ped_disk_set_partition_geom (disk, c_part, constraint,  partition.sector_start, partition.sector_end ) )
	{
		ped_constraint_destroy (constraint);
		ped_file_system_close (fs);
		return false ;
	}
	
	if ( ! ped_file_system_resize (fs, & c_part->geom, timer) )
	{
		ped_constraint_destroy (constraint);
		ped_file_system_close (fs);
		return false ;
	}
	
	ped_constraint_destroy (constraint);
	ped_file_system_close (fs);
	
	return ped_disk_commit (disk);
	
}

bool Device::Convert_Partition_Filesystem( const Partition & new_partition, PedTimer * timer)
{
	PedPartition *c_part =  NULL;
	PedFileSystemType* fs_type = NULL ;
	PedFileSystem *fs = NULL;
	
	c_part = ped_disk_get_partition( disk, new_partition .partition_number ) ;
	if ( ! c_part )
		return false;
		
	fs_type = ped_file_system_type_get ( new_partition .filesystem .c_str()   ) ;
	if ( ! fs_type )
		return false;
		
	fs = ped_file_system_create ( & c_part ->geom, fs_type, timer);
	if ( ! fs )
		return false;
	
	ped_file_system_close (fs);
	
	if ( ! ped_partition_set_system (c_part, fs_type ) )
		return false;
	
	return ped_disk_commit (disk) ;
}


bool Device::Resize_Extended( const Partition & partition , PedTimer *timer) 
{
	PedPartition *c_part =  NULL;
	
	c_part = ped_disk_extended_partition( disk ) ;
	if ( ! c_part )
		return false;
	
	PedConstraint *constraint = NULL;
	
	constraint = ped_constraint_any ( device );
	if ( ! constraint )
		return false;

	
	if ( ! ped_disk_set_partition_geom (disk, c_part, constraint,  partition.sector_start, partition.sector_end ) )
	{
		ped_constraint_destroy (constraint);
		return false ;
	}
	
	ped_partition_set_system ( c_part, NULL);
	
	ped_constraint_destroy (constraint);
		
	return ped_disk_commit (disk);
}

std::vector<Partition> Device::Get_Partitions()
{
	return device_partitions;
}

Sector Device::Get_Length() 
{
	return length ;
}

long Device::Get_Heads() 
{
	return heads ;
}

long Device::Get_Sectors() 
{
	return sectors ;
}

long Device:: Get_Cylinders() 
{
	return cylinders ;
}

Glib::ustring Device::Get_Model() 
{
	return model ;
}

Glib::ustring Device::Get_Path() 
{
	return path ;
}

Glib::ustring Device::Get_RealPath() 
{
	return realpath ;
}

Glib::ustring Device::Get_DiskType() 
{
	return disktype ;
}

bool Device::Get_any_busy() 
{
	for ( unsigned int t=0;t<device_partitions.size(); t++ )
		if ( device_partitions[t].busy )
			return true;
		
	return false;
}

Sector Device::Get_Used_Sectors( PedPartition *c_partition, const Glib::ustring & sym_path)
{
	/* This is quite an unreliable process, atm i try two different methods, but since both are far from perfect the results are 
	 * questionable. 
	 * - first i try geometry.get_used() in libpartedpp, i implemented this function to check the minimal size when resizing a partition. Disadvantage 
     *   of this method is the fact it won't work on mounted filesystems. Besides that, its SLOW
	 * - if the former method fails ( result is -1 ) i'll try to read the output from the df command ( 	df -k --sync <partition path> )
	 * - if this fails the filesystem on the partition is ( more or less ) unknown to the operating system and therefore the unused sectors cannot be calcualted
	 * - as soon as i have my internetconnection back i should ask people with more experience on this stuff for advice !
	*/
		
	//check if there is a (known) filesystem
	if ( ! c_partition ->fs_type )
		return -1;
	
	//used sectors are not relevant for swapspace
	if ( (Glib::ustring) c_partition ->fs_type ->name == "linux-swap" )
		return -1;
	
	//METHOD #1           
	//the getused method doesn't work on mounted partitions and for some filesystems ( i *guess* this is called check by andrew )
	if ( ! ped_partition_is_busy( c_partition ) && (Glib::ustring) c_partition ->fs_type ->name != "ntfs" )
	{ 		
		PedFileSystem *fs = NULL;
		PedConstraint *constraint = NULL;
		
		fs = ped_file_system_open( & c_partition ->geom ); //opening a filesystem is *SLOOOWW* :-(
		
		if ( fs )
		{	
			constraint = ped_file_system_get_resize_constraint (fs);
			ped_file_system_close (fs);
			if ( constraint && constraint->min_size != -1 ) //apperantly the compiler evaluates constraint != NULL before the min_size check. Very nice! :D
				return constraint->min_size ;
		}
		else 
			this ->error = error_message ;

	}
		
	//METHOD #2
	//this ony works for mounted ( and therefore known to the OS ) filesystems. My method is quite crude, keep in mind it's only temporarely ;-)
	Glib::ustring buf;
	system( ("df -k --sync " + sym_path + " | grep " + sym_path + " > /tmp/.tmp_gparted").c_str() );
	std::ifstream file_input( "/tmp/.tmp_gparted" );
	
	file_input >> buf; //skip first value
	file_input >> buf;
	if ( buf != "0" && ! buf.empty() ) 
	{ 
		file_input >> buf;
		file_input.close(); system( "rm -f /tmp/.tmp_gparted" );		
		return atoi( buf.c_str() ) * 1024/512 ;
	}
	file_input.close(); system( "rm -f /tmp/.tmp_gparted" );
	
	
	return -1 ; //all methods were unsuccesfull

}

Glib::ustring Device::Get_Flags( PedPartition *c_partition )
{
	temp = "";
		
	for ( unsigned short t = 0; t < flags.size()  ; t++ )
		if ( ped_partition_get_flag ( c_partition,  flags[ t ] ) )
			temp += (Glib::ustring) ped_partition_flag_get_name ( flags[ t ] ) + " ";
		
	return temp ;
}

bool Device::open_device_and_disk() 
{
	device = ped_device_get( this->realpath .c_str() );
	if ( device )
		disk 	= 	ped_disk_new( device );
	
	if ( ! device || ! disk  )
	{
		if ( device ) 
			{ ped_device_destroy( device ) ;	device = NULL ; }
				
		return false;
	}
	
	return true ;
}

void Device::close_device_and_disk()
{
	if ( device )
	{
		ped_device_destroy( device ) ;
		device = NULL ;
	}
	
	if ( disk )
	{
		ped_disk_destroy( disk ) ;
		disk = NULL ;
	}
}

Device::~Device()
{
	close_device_and_disk() ;	
}


} //GParted
